""" GenericSetup export / import support for smart folders
"""

import sys
from csv import reader
from csv import writer
from StringIO import StringIO
from zope.interface import implements
from xml.dom.minidom import parseString

from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse

from zExceptions import BadRequest

from Products.GenericSetup.interfaces import IFilesystemExporter
from Products.GenericSetup.interfaces import IFilesystemImporter
from Products.GenericSetup.interfaces import IDAVAware
from Products.GenericSetup.interfaces import ISetupTool
from Products.GenericSetup.content import FolderishExporterImporter, _globtest
from Products.CMFCore.utils import getToolByName

try:
    from Products.GenericSetup.utils import PageTemplateResource
except ImportError: # BBB:  no egg support
    from Products.PageTemplates.PageTemplateFile \
        import PageTemplateFile as PageTemplateResource

class SmartFolderExporterImporter(FolderishExporterImporter):
    """ Our hack to make Topics export their metadata as a file
    """

    implements(IFilesystemExporter, IFilesystemImporter)

    encoding = None
    _FILENAME = 'criteria.xml'
    _ROOT_TAGNAME = 'criteria'
    _SYNDFILENAME = 'syndication.xml'
    _SYNDROOT_TAGNAME = 'syndication'

    def __init__(self, context):
        self.context = context

    def listExportableItems(self):
        """ See IFilesystemExporter.
        """
        ###  XXX  Need to change to just exclude criterions
        criterions = self.context.listCriteria()
        criterion_ids = [x.getId() for x in criterions] 
        exclusions = ['syndication_information']
        return [x for x in FolderishExporterImporter.listExportableItems(self)
                   if x[0] not in criterion_ids and x[0] not in exclusions]

    def export(self, export_context, subdir, root=False):
        """ See IFilesystemExporter.
        """
        context = self.context

        if not root:
            subdir = '%s/%s' % (subdir, context.getId())

        exportable = self.listExportableItems()

        stream = StringIO()
        csv_writer = writer(stream)

        for object_id, object, adapter in exportable:
            csv_writer.writerow((object_id, object.getPortalTypeName()))

        export_context.writeDataFile('.objects',
                                    text=stream.getvalue(),
                                    content_type='text/comma-separated-values',
                                    subdir=subdir,
                                   )

        prop_adapter = IDAVAware(context, None)

        if prop_adapter is not None:
            export_context.writeDataFile('.topic_props',
                                         text=prop_adapter.manage_FTPget(),
                                         content_type='text/plain',
                                         subdir=subdir,
                                        )

        for object_id, object, adapter in exportable:
            if adapter is not None:
                adapter.export(export_context, subdir)

        template = PageTemplateResource('xml/%s' % self._FILENAME,
                                        globals()).__of__(self.context)
        export_context.writeDataFile(self._FILENAME,
                                     template(info=self._getExportInfo()),
                                     'text/xml',
                                     subdir,
                                    )
                                    
        # we export the syndication
        template = PageTemplateResource('xml/%s' % 'syndication.xml',
                                        globals()).__of__(self.context)
        export_context.writeDataFile(self._SYNDFILENAME,
                                     template(info=self._getExportSynd()),
                                     'text/xml',
                                     subdir,
                                    )
                                    
    
    def import_(self, import_context, subdir, root=False):
        """ See IFilesystemImporter.
        """
        context = self.context

        if not root:
            subdir = '%s/%s' % (subdir, context.getId())

        objects = import_context.readDataFile('.objects', subdir)
        if objects is None:
            return

        dialect = 'excel'
        stream = StringIO(objects)
        rowiter = reader(stream, dialect)
        ours = tuple(rowiter)
        our_ids = set([item[0] for item in ours])

        prior = set(context.contentIds())

        preserve = import_context.readDataFile('.preserve', subdir)
        if not preserve:
            preserve = set()
        else:
            preservable = prior.intersection(our_ids)
            preserve = set(_globtest(preserve, preservable))

        delete = import_context.readDataFile('.delete', subdir)
        if not delete:
            delete= set()
        else:
            deletable = prior.difference(our_ids)
            delete = set(_globtest(delete, deletable))

        # if it's in our_ids and NOT in preserve, or if it's not in
        # our_ids but IS in delete, we're gonna delete it
        delete = our_ids.difference(preserve).union(delete)

        for id in prior.intersection(delete):
            context._delObject(id)

        existing = context.objectIds()

        for object_id, portal_type in ours:

            if object_id not in existing:
                object = self._makeInstance(object_id, portal_type,
                                            subdir, import_context)
                if object is None:
                    logger = import_context.getLogger('SFWA')
                    logger.warning("Couldn't make instance: %s/%s" %
                                   (subdir, object_id))
                    continue

            wrapped = context._getOb(object_id)

            IFilesystemImporter(wrapped).import_(import_context, subdir)

        # set our props using our own dav props file
        properties = import_context.readDataFile('.topic_props', subdir)

        prop_adapter = IDAVAware(context, None)
        
        if properties is not None:
            prop_file = StringIO()
            prop_file.write(properties)
            prop_file.flush()
            prop_file.seek(0)
            resp = HTTPResponse(stdout=None)
            environ = {}
            environ['SERVER_NAME']='foo'
            environ['SERVER_PORT']='80'
            environ['REQUEST_METHOD'] = 'PUT'
            req = HTTPRequest(stdin=prop_file, environ=environ, response=resp)
            req.processInputs()

            prop_adapter.PUT(req, None)
        
        # create criteria
        self.encoding = import_context.getEncoding()

        if import_context.shouldPurge():
            self._purgeContext()
        
        data = import_context.readDataFile(self._FILENAME,
                                           subdir)
        if data is not None:
            dom = parseString(data)
            root = dom.firstChild
            assert root.tagName == self._ROOT_TAGNAME
            self._updateFromDOM(root)
            
        # create syndication
        if import_context.shouldPurge():
            self._purgeContext()

        data = import_context.readDataFile(self._SYNDFILENAME,
                                           subdir)
        if data is not None:
            dom = parseString(data)
            root = dom.firstChild
            assert root.tagName == self._SYNDROOT_TAGNAME
            self._updateSyndFromDOM(root)

    def _makeInstance(self, id, portal_type, subdir, import_context):

        context = self.context
        
        tool = getToolByName(context, 'portal_types')

        try:
            tool.constructContent(portal_type, context, id)
        except ValueError: # invalid type
            return None

        content = context._getOb(id)

        return content

    def _getNodeAttr(self, node, attrname, default=None):
        attr = node.attributes.get(attrname)
        if attr is None:
            return default
        value = attr.value
        if isinstance(value, unicode) and self.encoding is not None:
            value = value.encode(self.encoding)
        return value

    def _purgeContext(self):
        context = self.context
        criterions = context.listCriteria()
        criterion_ids = [x.getId() for x in criterions]
        for criterion_id in criterion_ids:
            self.context._delObject(criterion_id)

    def _updateFromDOM(self, root):
        for criterion in root.getElementsByTagName('criterion'):
            c_type = self._getNodeAttr(criterion, 'type', None)
            field = self._getNodeAttr(criterion, 'field', None)
            criterion_id = self._getNodeAttr(criterion, 'criterion_id', None)
            attributes = {}
            for attribute in criterion.getElementsByTagName('attribute'):
                name = self._getNodeAttr(attribute, 'name', None)
                value = self._getNodeAttr(attribute, 'value', None)
                if name == 'reversed':
                    value = value in ('True', 'true', '1')
                if type(value) in (str,):
                    if ',' in value:
                        # we have a list of values
                        new_value = value.split(',')
                        value = tuple(new_value)
                attributes[name] = value
            
            try:
                self.context.addCriterion(field, c_type)
            except BadRequest:# remove me
                pass # the criterion already exist
            added = self.context.getCriterion(criterion_id)
            added.edit(**attributes)
            
    def _updateSyndFromDOM(self, root):
        ps = getToolByName(self.context, 'portal_syndication')
        for param in root.getElementsByTagName('param'):
            update_period = self._getNodeAttr(param, 'update_period', None)
            update_freq = self._getNodeAttr(param, 'update_freq', None)
            update_base = self._getNodeAttr(param, 'update_base', None)
            max_items = self._getNodeAttr(param, 'max_items', None)

            ps.editSyInformationProperties(self.context, 
                                            updatePeriod=update_period, 
                                            updateFrequency=update_freq, 
                                            updateBase=update_base, 
                                            max_items=max_items)

    def _getExportInfo(self):
        context = self.context
        criterion_info = []
        criterions = context.listCriteria()
        criterions = [(x.getId(), x) for x in criterions]

        for criterion_id, criterion in criterions:

            info = {'criterion_id': criterion_id,
                    'type': criterion.meta_type,
                    'field': criterion.Field(),
                    'attributes': []
                   }

            attributes = info['attributes']

            for field in criterion.schema.fields():
                value = field.get(criterion)
                if type(value) in (tuple, list):
                    value = ','.join(value)
                attributes.append((field.getName(), value))

            criterion_info.append(info)

        return {'criteria': criterion_info,
               }
               
    def _getExportSynd(self):
        context = self.context
        ps = getToolByName(context, 'portal_syndication')
        update_period = ps.getUpdatePeriod(context)
        update_freq = ps.getUpdateFrequency(context)
        update_base = ps.getUpdateBase(context)
        max_items = ps.getMaxItems(context)
        
        syndication_info = []

        info = {'update_period': update_period,
                'update_freq': update_freq,
                'update_base': update_base,
                'max_items': max_items
               }

        syndication_info.append(info)

        return {'syndication_param': syndication_info,
               }
